#include "adevs.h"
#include <iostream>
#include <cassert>
#include <cstdlib>
#include <sys/time.h>
using namespace std;
using namespace adevs;

/* A model of a person. The lifecycle stages are born -> midlife -> end of life -> death. */
class Person:
	public Model<int>
{
	public:
		Person(string name):m_name(name),stage(0){}
		Time init(SimEnv<int>* env)
		{
			cout << name() << " is born @ " << env->now().real
				<< "," << env->now().logical << endl;
			/* The person will be midlife in 20 to 60 years. The
			   time advance must be added to now() and not the
			   other way round. Recall that + does not commute. */
			return env->now()+Time(20+rand()%40,0);
		}
		/* This is called twice. Once when the person reaches midlife
		   and again just prior to death. */
		Time update(SimEnv<int>* env)
		{
			/* Advance the life stage */
			stage++;
			cout << name() << " is " << ((stage==1) ? "midlife" : "end of life")
				<< " @ " << env->now().real << "," <<
				env->now().logical << endl;
			/* Advance to end of life in 0 to 40 years plus some number of
			   logical time increments. Notice that this will always move
			   us into the future. */
			if (stage == 1)
				return env->now()+Time(rand()%41,rand()%12+1);
			/* We have reached the end. Return an infinite time of
			   next event and remove the model. */
			env->remove(this);
			return adevs_inf<Time>();
		}
		/* This method is never called because nobody sends us messages. */
		Time update(SimEnv<int>*,vector<int>&) { return adevs_inf<Time>(); }
		/* This is called once at the end of our life. */
		void fini(Time now)
		{
			cout << name() << " has died @ " <<
				now.real << "," << now.logical << endl;
			/* Delete ourselves. This is not necessary, but depends
			   on your particular memory management strategy for
			   remove models. */
			delete this;
		}
		/* Overriden from the base class. */
		string name() { return m_name; }
	private:
		/* Our name */
		const std::string m_name;
		/* The stage of life, starts at zero and ends at two */
		int stage;
};

int main()
{
	/* Seed the random number generator */
	srand(time(NULL));
	/* Create the model */
	Person* person = new Person("George");
	/* Create the simulation context */
	Simulator<int> sim;
	/* Add the model to the context */
	sim.add(person);
	/* Run until the next event is at infinity */
	while (sim.next_event_time() < adevs_inf<Time>())
		sim.exec_next_event();
	/* Person destroyed itself and the simulator will be
	   deleted when it goes out of context. */
	return 0;
}
